using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using SDL2;
using SlatedGameToolkit.Framework.DataTypes;
using SlatedGameToolkit.Framework.Exceptions;

namespace SlatedGameToolkit.Framework.Graphics.Window
{
    public delegate WindowRegion WindowRegionHit(FloatVector2 hitPoint);
    public sealed class WindowHandle : IDisposable
    {
        public readonly IntPtr window;
        public event WindowRegionHit windowRegionHitEvent;
        IntPtr windowSurfaceHandle;

        public bool Shown {
            set {
                if (value) {
                    SDL.SDL_ShowWindow(window);
                } else {
                    SDL.SDL_HideWindow(window);
                }
            }

            get {
                return ((SDL.SDL_WindowFlags) Enum.Parse(typeof(SDL.SDL_WindowFlags), SDL.SDL_GetWindowFlags(window).ToString())).HasFlag(SDL.SDL_WindowFlags.SDL_WINDOW_SHOWN);
            }
        }

        public string WindowTitle {
            set {
                SDL.SDL_SetWindowTitle(window, value);
            }

            get {
                return SDL.SDL_GetWindowTitle(window);
            }
        }

        public bool WindowBordered {
            set {
                SDL.SDL_SetWindowBordered(window, value ? SDL.SDL_bool.SDL_TRUE : SDL.SDL_bool.SDL_FALSE);
            }

            get {
                int top, bottom, left, right;
                int errorCode = SDL.SDL_GetWindowBordersSize(window, out top, out left, out bottom, out right);
                if (errorCode < 0) throw new OptionalSDLException();
                return top > 0 || bottom > 0 || left > 0 || right > 0;
            }
        }

        public bool WindowResizable {
            set {
                SDL.SDL_SetWindowResizable(window, value ? SDL.SDL_bool.SDL_TRUE : SDL.SDL_bool.SDL_FALSE);
            }
            get {
                return ((SDL.SDL_WindowFlags) Enum.Parse(typeof(SDL.SDL_WindowFlags), SDL.SDL_GetWindowFlags(window).ToString())).HasFlag(SDL.SDL_WindowFlags.SDL_WINDOW_RESIZABLE);
            }
        }

        public FloatRectangle WindowBoundaries {
            set {
                SDL.SDL_SetWindowPosition(window, (int)value.X1, (int)value.Y1);
                SDL.SDL_SetWindowSize(window, (int)value.Width, (int)value.Height);
            }
            get {
                int x, y, width, height;
                SDL.SDL_GetWindowSize(window, out x, out y);
                SDL.SDL_GetWindowPosition(window, out width, out height);
                FloatRectangle rectangle = new FloatRectangle();
                rectangle.X1 = x;
                rectangle.Y2 = y;
                rectangle.Width = width;
                rectangle.Height = height;

                return rectangle;
            }
        }

        public FloatVector2 MaximumSize {
            set {
                SDL.SDL_SetWindowMaximumSize(window, (int)value.x, (int)value.y);
            }
            get {
                int width, height;
                SDL.SDL_GetWindowMaximumSize(window, out width, out height);
                FloatVector2 maxSize = new FloatVector2();
                maxSize.x = width;
                maxSize.y = height;
                return maxSize;
            }
        }

        public float Opacity {
            set {
                int errorCode = SDL.SDL_SetWindowOpacity(window, value);
                if (errorCode < 0) throw new OptionalSDLException();
            }
            get {
                float value;
                int errorCode = SDL.SDL_GetWindowOpacity(window, out value);
                if (errorCode < 0) throw new OptionalSDLException();
                return value;
            }
        }

        public FloatVector2 MinimumSize {
            set {
                SDL.SDL_SetWindowMinimumSize(window, (int)value.x, (int)value.y);
            }
            get {
                int width, height;
                SDL.SDL_GetWindowMinimumSize(window, out width, out height);
                FloatVector2 maxSize = new FloatVector2();
                maxSize.x = width;
                maxSize.y = height;
                return maxSize;
            }
        }

        public bool GrabbingInput {
            set {
                SDL.SDL_SetWindowGrab(window, value ? SDL.SDL_bool.SDL_TRUE : SDL.SDL_bool.SDL_FALSE);
            }
            get {
                return SDL.SDL_GetWindowGrab(window) == SDL.SDL_bool.SDL_TRUE ? true : false;
            }
        }

        public WindowHandle(string title, int x, int y, int width, int height, bool specialRegions, WindowOptions options) {
            window = SDL.SDL_CreateWindow(title, x, y, width, height, SDL.SDL_WindowFlags.SDL_WINDOW_OPENGL | options.ToSDLWindowFlag());
            if (window == null) {
                throw new FrameworkSDLException();
            }
            if (SDL.SDL_SetWindowHitTest(window, SpecialRegionHit, IntPtr.Zero) < 0) {
                throw new OptionalSDLException();
            }
            windowSurfaceHandle = SDL.SDL_GetWindowSurface(window);
        }

        private SDL.SDL_HitTestResult SpecialRegionHit(IntPtr window, IntPtr hitPtr, IntPtr data) {
            SDL.SDL_Point SDLPoint = Marshal.PtrToStructure<SDL.SDL_Point>(hitPtr);
            FloatVector2 point = new FloatVector2(SDLPoint.x, SDLPoint.y);
            WindowRegion region = windowRegionHitEvent.Invoke(point);
            return region.ToSDLHitTestResult();
        }

        public void RaiseToTop() {
            SDL.SDL_RaiseWindow(window);
        }

        public int GetDisplayIndex() {
            int index = SDL.SDL_GetWindowDisplayIndex(window);
            if (index < 0) throw new FrameworkSDLException();
            return index;
        }

        public void Dispose()
        {
            SDL.SDL_FreeSurface(windowSurfaceHandle);
            SDL.SDL_DestroyWindow(window);
        }

        ~WindowHandle() {
            Dispose();
        }
    }
}